

set -ex



pushd test-cmake
cmake . -GNinja -DCMAKE_BUILD_TYPE=Release
cmake --build . --verbose
popd
test -f ${CONDA_PREFIX}/lib/libopencv_aruco.so
test -f ${CONDA_PREFIX}/lib/libopencv_bgsegm.so
test -f ${CONDA_PREFIX}/lib/libopencv_calib3d.so
test -f ${CONDA_PREFIX}/lib/libopencv_ccalib.so
test -f ${CONDA_PREFIX}/lib/libopencv_core.so
test -f ${CONDA_PREFIX}/lib/libopencv_datasets.so
test -f ${CONDA_PREFIX}/lib/libopencv_features2d.so
test -f ${CONDA_PREFIX}/lib/libopencv_flann.so
test -f ${CONDA_PREFIX}/lib/libopencv_fuzzy.so
test -f ${CONDA_PREFIX}/lib/libopencv_highgui.so
test -f ${CONDA_PREFIX}/lib/libopencv_imgcodecs.so
test -f ${CONDA_PREFIX}/lib/libopencv_imgproc.so
test -f ${CONDA_PREFIX}/lib/libopencv_line_descriptor.so
test -f ${CONDA_PREFIX}/lib/libopencv_ml.so
test -f ${CONDA_PREFIX}/lib/libopencv_optflow.so
test -f ${CONDA_PREFIX}/lib/libopencv_phase_unwrapping.so
test -f ${CONDA_PREFIX}/lib/libopencv_photo.so
test -f ${CONDA_PREFIX}/lib/libopencv_plot.so
test -f ${CONDA_PREFIX}/lib/libopencv_reg.so
test -f ${CONDA_PREFIX}/lib/libopencv_rgbd.so
test -f ${CONDA_PREFIX}/lib/libopencv_saliency.so
test -f ${CONDA_PREFIX}/lib/libopencv_shape.so
test -f ${CONDA_PREFIX}/lib/libopencv_stereo.so
test -f ${CONDA_PREFIX}/lib/libopencv_stitching.so
test -f ${CONDA_PREFIX}/lib/libopencv_structured_light.so
test -f ${CONDA_PREFIX}/lib/libopencv_superres.so
test -f ${CONDA_PREFIX}/lib/libopencv_surface_matching.so
test -f ${CONDA_PREFIX}/lib/libopencv_tracking.so
test -f ${CONDA_PREFIX}/lib/libopencv_video.so
test -f ${CONDA_PREFIX}/lib/libopencv_videoio.so
test -f ${CONDA_PREFIX}/lib/libopencv_videostab.so
test -f ${CONDA_PREFIX}/lib/libopencv_xfeatures2d.so
test -f ${CONDA_PREFIX}/lib/libopencv_ximgproc.so
test -f ${CONDA_PREFIX}/lib/libopencv_xphoto.so
exit 0
